﻿using System;
using System.Collections.Generic;
using UnityEditorInternal;
using UnityEngine;
using UnityObject = UnityEngine.Object;

namespace CloneGames.ObjectPool
{
    internal static class ObjectPoolServiceInjector
    {
        [RuntimeInitializeOnLoadMethod]
        public static void Inject()
        {
            var go = new GameObject("[CG Object Pool]");
            ObjectPoolService.__instance = go.AddComponent<ObjectPoolService>();
            UnityObject.DontDestroyOnLoad(go);
        }
    }
    
    public class ObjectPoolService : MonoBehaviour
    {
        internal static ObjectPoolService __instance;

        public static ObjectPoolService Instance => __instance;
        
        private Transform pooledObjectRoot;
        private Dictionary<UnityObject, List<UnityObject>> pooledObjects = new Dictionary<UnityObject, List<UnityObject>>();
        private Dictionary<UnityObject, UnityObject> instanceKey = new Dictionary<UnityObject, UnityObject>();

        private void Awake()
        {
            pooledObjectRoot = new GameObject("Pooled Objects").transform;
            pooledObjectRoot.gameObject.SetActive(false);
            pooledObjectRoot.SetParent(transform);
        }

        public T Spawn<T>(T original) where T : UnityObject
        {
            // If there are any pooled instances, return one
            if (pooledObjects.TryGetValue(original, out var list) && list.Count > 0)
            {
                var instance = list[0];
                list.RemoveAt(0);
                SetTransform(instance, null);
                return instance as T;
            }
            
            var newInstance = Instantiate(original);
            RegisterInstance(newInstance, original);
            return newInstance;
        }

        public T Spawn<T>(T original, Transform parent) where T : UnityObject
        {
            // If there are any pooled instances, return one
            if (pooledObjects.TryGetValue(original, out var list) && list.Count > 0)
            {
                var instance = list[0];
                list.RemoveAt(0);
                SetTransform(instance, parent);
                return instance as T;
            }
            
            var newInstance = Instantiate(original, parent);
            RegisterInstance(newInstance, original);
            return newInstance;
        }

        public T Spawn<T>(T original, Vector3 position, Quaternion rotation) where T : UnityObject
        {
            // If there are any pooled instances, return one
            if (pooledObjects.TryGetValue(original, out var list) && list.Count > 0)
            {
                var instance = list[0];
                list.RemoveAt(0);
                SetTransform(instance, position, rotation, null);
                return instance as T;
            }
            
            var newInstance = Instantiate(original, position, rotation);
            RegisterInstance(newInstance, original);
            return newInstance;
        }

        public T Spawn<T>(T original, Vector3 position, Quaternion rotation, Transform parent) where T : UnityObject
        {
            // If there are any pooled instances, return one
            if (pooledObjects.TryGetValue(original, out var list) && list.Count > 0)
            {
                var instance = list[0];
                list.RemoveAt(0);
                SetTransform(instance, position, rotation, parent);
                return instance as T;
            }
            
            var newInstance = Instantiate(original, position, rotation, parent);
            RegisterInstance(newInstance, original);
            return newInstance;
        }

        public void Recycle(UnityObject instance)
        {
            if (instanceKey.TryGetValue(instance, out var key))
            {
                if (pooledObjects.TryGetValue(key, out var pool))
                {
                    pool.Add(instance);
                }
                else
                {
                    pooledObjects.Add(key, new List<UnityObject>
                    {
                        instance
                    });
                }

                switch (instance)
                {
                    case Transform t:
                        t.SetParent(pooledObjectRoot);
                        break;
                        
                    case GameObject go:
                        go.transform.SetParent(pooledObjectRoot);
                        break;
                        
                    case Component comp:
                        comp.transform.SetParent(pooledObjectRoot);
                        break;
                }
            }
        }

        public void ClearPool<T>(T original) where T : UnityObject
        {
            if (pooledObjects.TryGetValue(original, out var pool))
            {
                pooledObjects.Remove(original);
                
                foreach (var entry in pool)
                {
                    instanceKey.Remove(entry);
                    
                    switch (entry)
                    {
                        case Component comp:
                            Destroy(comp);
                            break;
                        
                        default:
                            Destroy(entry);
                            break;
                    }
                }
            }
        }

        private void SetTransform(UnityObject obj, Transform parent)
        {
            switch (obj)
            {
                case Component comp:
                    comp.transform.SetParent(parent);
                    break;
                
                case GameObject go:
                    go.transform.SetParent(parent);
                    break;
            }
        }

        private void SetTransform(UnityObject obj, Vector3 position, Quaternion rotation)
        {
            switch (obj)
            {
                case Component comp:
                    comp.transform.position = position;
                    comp.transform.rotation = rotation;
                    break;
                
                case GameObject go:
                    go.transform.position = position;
                    go.transform.rotation = rotation;
                    break;
            }
        }

        private void SetTransform(UnityObject obj, Vector3 position, Quaternion rotation, Transform parent)
        {
            switch (obj)
            {
                case Component comp:
                    comp.transform.position = position;
                    comp.transform.rotation = rotation;
                    comp.transform.SetParent(parent);
                    break;
                
                case GameObject go:
                    go.transform.position = position;
                    go.transform.rotation = rotation;
                    go.transform.SetParent(parent);
                    break;
            }
        }

        private void RegisterInstance(UnityObject instance, UnityObject key)
        {
            if (!instanceKey.TryGetValue(instance, out _))
            {
                instanceKey.Add(instance, key);
            }
        }
    }
}